package net.sourceforge.nattable.export.excel;

import static net.sourceforge.nattable.util.ObjectUtils.isNotNull;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import net.sourceforge.nattable.config.IConfigRegistry;
import net.sourceforge.nattable.export.FileOutputStreamProvider;
import net.sourceforge.nattable.export.ILayerExporter;
import net.sourceforge.nattable.export.IOutputStreamProvider;
import net.sourceforge.nattable.layer.cell.LayerCell;
import net.sourceforge.nattable.style.CellStyleAttributes;
import net.sourceforge.nattable.style.CellStyleProxy;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.widgets.Shell;

public class ExcelExporter implements ILayerExporter {

	private static final String EXCEL_HEADER_FILE = "excelExportHeader.txt";
	
	private final IOutputStreamProvider outputStreamProvider;

	public ExcelExporter() {
		this(new FileOutputStreamProvider("table_export.xls", new String[] { "Excel Workbok (*.xls)" }, new String[] { "*.xls" }));
	}
	
	public ExcelExporter(IOutputStreamProvider outputStreamProvider) {
		this.outputStreamProvider = outputStreamProvider;
	}
	
	public OutputStream getOutputStream(Shell shell) {
		return outputStreamProvider.getOutputStream(shell);
	}
	
	public void exportBegin(OutputStream outputStream) throws IOException {
	}

	public void exportEnd(OutputStream outputStream) throws IOException {
	}

	public void exportLayerBegin(OutputStream outputStream, String layerName) throws IOException {
		writeHeader(outputStream);
		outputStream.write(asBytes("<body><table border='1'>"));
	}

	private void writeHeader(OutputStream outputStream) throws IOException {
		InputStream headerStream = null;
		try {
			headerStream = this.getClass().getResourceAsStream(EXCEL_HEADER_FILE);
			int c;
			while ((c = headerStream.read()) != -1) {
				outputStream.write(c);
			}
		} catch (Exception e) {
			logError(e);
		} finally {
			if (isNotNull(headerStream)) {
				headerStream.close();
			}
		}
	}

	public void exportLayerEnd(OutputStream outputStream, String layerName) throws IOException {
		outputStream.write(asBytes("</table></body></html>"));
	}

	public void exportRowBegin(OutputStream outputStream, int rowPosition) throws IOException {
		outputStream.write(asBytes("<tr>\n"));
	}

	public void exportRowEnd(OutputStream outputStream, int rowPosition) throws IOException {
		outputStream.write(asBytes("</tr>\n"));
	}

	public void exportCell(OutputStream outputStream, Object exportDisplayValue, LayerCell cell, IConfigRegistry configRegistry) throws IOException {
		CellStyleProxy cellStyle = new CellStyleProxy(configRegistry, cell.getDisplayMode(), cell.getConfigLabels().getLabels());
		Color fg = cellStyle.getAttributeValue(CellStyleAttributes.FOREGROUND_COLOR);
		Color bg = cellStyle.getAttributeValue(CellStyleAttributes.BACKGROUND_COLOR);
		Font font = cellStyle.getAttributeValue(CellStyleAttributes.FONT);

		String htmlAttributes = String.format("style='color: %s; background-color: %s; %s;'",
		                     getColorInCSSFormat(fg),
		                     getColorInCSSFormat(bg),
		                     getFontInCSSFormat(font));
		
		String htmlText = exportDisplayValue != null ? exportDisplayValue.toString() : "";
		
		if (htmlText.startsWith(" ")) {
			htmlAttributes += " x:str=\"'" + htmlText + "\";";
			htmlText = htmlText.replaceFirst("^(\\ *)", "<span style='mso-spacerun:yes'>$1</span>");
		}
		
		outputStream.write(asBytes(String.format("\t<td %s>%s</td>\n", htmlAttributes, htmlText)));
	}

	private byte[] asBytes(String string) {
		return string.getBytes();
	}

	private void logError(Exception e) {
		System.err.println("Excel Exporter failed: " + e.getMessage());
		e.printStackTrace(System.err);
	}
	
	private String getFontInCSSFormat(Font font) {
		FontData fontData = font.getFontData()[0];
		String fontName = fontData.getName();
		int fontStyle = fontData.getStyle();
		String HTML_STYLES[] = new String[] { "NORMAL", "BOLD", "ITALIC" };

		return String.format("font: %s; font-family: %s",
		                     fontStyle <= 2 ? HTML_STYLES[fontStyle] : HTML_STYLES[0],
		                     fontName);
	}

	private String getColorInCSSFormat(Color color) {
		return String.format("rgb(%d,%d,%d)",
		                     Integer.valueOf(color.getRed()),
		                     Integer.valueOf(color.getGreen()),
		                     Integer.valueOf(color.getBlue()));
	}
	
}
