package net.sourceforge.nattable.filterrow;

import static net.sourceforge.nattable.filterrow.FilterRowDataLayer.FILTER_ROW_COLUMN_LABEL_PREFIX;
import static net.sourceforge.nattable.filterrow.config.FilterRowConfigAttributes.FILTER_DISPLAY_CONVERTER;
import static net.sourceforge.nattable.style.DisplayMode.NORMAL;
import static net.sourceforge.nattable.util.ObjectUtils.isEmpty;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import net.sourceforge.nattable.config.IConfigRegistry;
import net.sourceforge.nattable.data.IDataProvider;
import net.sourceforge.nattable.data.convert.IDisplayConverter;
import net.sourceforge.nattable.filterrow.event.FilterAppliedEvent;
import net.sourceforge.nattable.layer.ILayer;
import net.sourceforge.nattable.persistence.IPersistable;
import net.sourceforge.nattable.util.ObjectUtils;
import net.sourceforge.nattable.util.PersistenceUtils;

/**
 * Data provider for the filter row<br/>
 * - Stores filter strings
 * - Applies them to the {@link MatcherEditor} on the {@link FilterList}
 */
public class FilterRowDataProvider<T> implements IDataProvider, IPersistable {

	private final IFilterStrategy<T> filterStrategy;
	private final ILayer columnHeaderLayer;
	private final IDataProvider columnHeaderDataProvider;
	private final IConfigRegistry configRegistry;
	
	private Map<Integer, Object> filterObjectByIndex = new HashMap<Integer, Object>();
	private int rowCount = 1;
	
	public FilterRowDataProvider(IFilterStrategy<T> filterStrategy, ILayer columnHeaderLayer, IDataProvider columnHeaderDataProvider, IConfigRegistry configRegistry) {
		this.filterStrategy = filterStrategy;
		this.columnHeaderLayer = columnHeaderLayer;
		this.columnHeaderDataProvider = columnHeaderDataProvider;
		this.configRegistry = configRegistry;
	}

	public int getColumnCount() {
		return columnHeaderDataProvider.getColumnCount();
	}

	public Object getDataValue(int columnIndex, int rowIndex) {
		return filterObjectByIndex.get(columnIndex);
	}

	public int getRowCount() {
		return rowCount;
	}

	public void setRowCount(int rowCount) {
		this.rowCount = rowCount;
	}

	public void setDataValue(int columnIndex, int rowIndex, Object newValue) {
		columnIndex = columnHeaderLayer.getColumnIndexByPosition(columnIndex);

		if (ObjectUtils.isNotNull(newValue)) {
			filterObjectByIndex.put(columnIndex, newValue);
		} else {
			filterObjectByIndex.remove(columnIndex);
		}

		filterStrategy.applyFilter(filterObjectByIndex);

		columnHeaderLayer.fireLayerEvent(new FilterAppliedEvent(columnHeaderLayer));
	}

	// Load/save state

	public void saveState(String prefix, Properties properties) {
		Map<Integer, String> filterTextByIndex = new HashMap<Integer, String>();
		for(Integer columnIndex : filterObjectByIndex.keySet()){
			final IDisplayConverter converter = configRegistry.getConfigAttribute(
					FILTER_DISPLAY_CONVERTER, NORMAL, FILTER_ROW_COLUMN_LABEL_PREFIX + columnIndex);
			filterTextByIndex.put(columnIndex, (String) converter.canonicalToDisplayValue(filterObjectByIndex.get(columnIndex)));
		}
		
		String string = PersistenceUtils.mapAsString(filterTextByIndex);

		if (!isEmpty(string)) {
			properties.put(prefix + FilterRowDataLayer.PERSISTENCE_KEY_FILTER_ROW_TOKENS, string);
		}
	}
	
	public void loadState(String prefix, Properties properties) {
		filterObjectByIndex.clear();
		
		try {
			Object property = properties.get(prefix + FilterRowDataLayer.PERSISTENCE_KEY_FILTER_ROW_TOKENS);
			Map<Integer, String> filterTextByIndex = PersistenceUtils.parseString(property);
			for (Integer columnIndex : filterTextByIndex.keySet()) {
				final IDisplayConverter converter = configRegistry.getConfigAttribute(
						FILTER_DISPLAY_CONVERTER, NORMAL, FILTER_ROW_COLUMN_LABEL_PREFIX + columnIndex);
				filterObjectByIndex.put(columnIndex, converter.displayToCanonicalValue(filterTextByIndex.get(columnIndex)));
			}
		} catch (Exception e) {
			System.err.println("Error while restoring filter row text: " + e.getMessage());
		}
		
		filterStrategy.applyFilter(filterObjectByIndex);
	}

	public void clearAllFilters() {
		filterObjectByIndex.clear();
		filterStrategy.applyFilter(filterObjectByIndex);
	}
	
}