package net.sourceforge.nattable.painter.cell.decorator;

import net.sourceforge.nattable.config.IConfigRegistry;
import net.sourceforge.nattable.layer.cell.LayerCell;
import net.sourceforge.nattable.painter.cell.ICellPainter;
import net.sourceforge.nattable.style.CellStyleAttributes;
import net.sourceforge.nattable.style.CellStyleUtil;
import net.sourceforge.nattable.ui.util.CellEdgeEnum;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.graphics.Region;

/**
 * Decorates a cell painter with another cell painter.
 */
public class CellPainterDecorator implements ICellPainter {

	private final ICellPainter baseCellPainter;
	private final CellEdgeEnum cellEdge;
	private final int spacing;
	private final ICellPainter decoratorCellPainter;
    private boolean interiorPainterToSpanFullWidth;

	public CellPainterDecorator(ICellPainter baseCellPainter, CellEdgeEnum cellEdge, ICellPainter decoratorCellPainter) {
		this(baseCellPainter, cellEdge, 2, decoratorCellPainter);
	}
	
	public CellPainterDecorator(ICellPainter baseCellPainter, CellEdgeEnum cellEdge, int spacing, ICellPainter decoratorCellPainter) {
		this.baseCellPainter = baseCellPainter;
		this.cellEdge = cellEdge;
		this.spacing = spacing;
		this.decoratorCellPainter = decoratorCellPainter;
	}

    public void setBaseCellPainterSpansWholeCell(boolean interiorPainterToSpanFullWidth) {
        this.interiorPainterToSpanFullWidth = interiorPainterToSpanFullWidth;
    }
	
	public int getPreferredWidth(LayerCell cell, GC gc, IConfigRegistry configRegistry) {
		switch (cellEdge) {
		case BOTTOM:
		case TOP:
		    return spacing
					+ Math.max(
							baseCellPainter.getPreferredWidth(cell, gc, configRegistry),
							decoratorCellPainter.getPreferredWidth(cell, gc, configRegistry)
					);
        default:
            break;
		}

		return baseCellPainter.getPreferredWidth(cell, gc, configRegistry)
				+ spacing
				+ decoratorCellPainter.getPreferredWidth(cell, gc, configRegistry);
	}

	public int getPreferredHeight(LayerCell cell, GC gc, IConfigRegistry configRegistry) {
		switch (cellEdge) {
		case LEFT:
		case RIGHT:
			return spacing
					+ Math.max(
							baseCellPainter.getPreferredHeight(cell, gc, configRegistry),
							decoratorCellPainter.getPreferredHeight(cell, gc, configRegistry)
					);
		default:
		    break;
		}

		return baseCellPainter.getPreferredHeight(cell, gc, configRegistry)
				+ spacing
				+ decoratorCellPainter.getPreferredHeight(cell, gc, configRegistry);
	}

	public void paintCell(LayerCell cell, GC gc, Rectangle adjustedCellBounds, IConfigRegistry configRegistry) {
		Rectangle baseCellPainterBounds = getBaseCellPainterBounds(cell, gc, adjustedCellBounds, configRegistry);
		Rectangle decoratorCellPainterBounds = getDecoratorCellPainterBounds(cell, gc, adjustedCellBounds, configRegistry);

		baseCellPainter.paintCell(cell, gc, baseCellPainterBounds, configRegistry);
		decoratorCellPainter.paintCell(cell, gc, decoratorCellPainterBounds, configRegistry);
		
		// now paint the spacing (if there is any)
		Region spacingRegion = new Region();
		spacingRegion.add(adjustedCellBounds);
		spacingRegion.subtract(baseCellPainterBounds);
		spacingRegion.subtract(decoratorCellPainterBounds);
		
		Color originalBg = gc.getBackground();
		gc.setBackground(CellStyleUtil.getCellStyle(cell, configRegistry).getAttributeValue(CellStyleAttributes.BACKGROUND_COLOR));
		
		gc.fillRectangle(spacingRegion.getBounds());
		
		gc.setBackground(originalBg);
		
		spacingRegion.dispose();
	}

	public Rectangle getBaseCellPainterBounds(LayerCell cell, GC gc, Rectangle adjustedCellBounds, IConfigRegistry configRegistry) {
		int preferredDecoratorWidth = decoratorCellPainter.getPreferredWidth(cell, gc, configRegistry);
		int preferredDecoratorHeight = decoratorCellPainter.getPreferredHeight(cell, gc, configRegistry);
		
		// grab any extra space:
		int grabbedPreferredWidth = adjustedCellBounds.width - (interiorPainterToSpanFullWidth ? 0 : preferredDecoratorWidth) - spacing;
		int grabbedPreferredHeight = adjustedCellBounds.height - (interiorPainterToSpanFullWidth ? 0 : preferredDecoratorHeight) - spacing;
		
		switch (cellEdge) {
		case LEFT:
			return new Rectangle(
					adjustedCellBounds.x + preferredDecoratorWidth + spacing,
					adjustedCellBounds.y,
					grabbedPreferredWidth,
					adjustedCellBounds.height
					).intersection(adjustedCellBounds);
		case RIGHT:
			return new Rectangle(
					adjustedCellBounds.x,
					adjustedCellBounds.y,
					grabbedPreferredWidth,
					adjustedCellBounds.height
					).intersection(adjustedCellBounds);
		case TOP:
			return new Rectangle(
					adjustedCellBounds.x,
					adjustedCellBounds.y + preferredDecoratorHeight + spacing,
					adjustedCellBounds.width,
					grabbedPreferredHeight
					).intersection(adjustedCellBounds);
		case BOTTOM:
			return new Rectangle(
					adjustedCellBounds.x,
					adjustedCellBounds.y,
					adjustedCellBounds.width,
					grabbedPreferredHeight
					).intersection(adjustedCellBounds);
		case NONE:
		    break;
		}
		
		return null;
	}

	public Rectangle getDecoratorCellPainterBounds(LayerCell cell, GC gc, Rectangle adjustedCellBounds, IConfigRegistry configRegistry) {
		int preferredDecoratorWidth = decoratorCellPainter.getPreferredWidth(cell, gc, configRegistry);
		int preferredDecoratorHeight = decoratorCellPainter.getPreferredHeight(cell, gc, configRegistry);
		
		switch (cellEdge) {
		case LEFT:
			return new Rectangle(
					adjustedCellBounds.x,
					adjustedCellBounds.y,
					preferredDecoratorWidth,
					adjustedCellBounds.height
					).intersection(adjustedCellBounds);
		case RIGHT:
			return new Rectangle(
					adjustedCellBounds.x + adjustedCellBounds.width - preferredDecoratorWidth,
					adjustedCellBounds.y,
					preferredDecoratorWidth,
					adjustedCellBounds.height
					).intersection(adjustedCellBounds);
		case TOP:
			return new Rectangle(
					adjustedCellBounds.x,
					adjustedCellBounds.y,
					adjustedCellBounds.width,
					preferredDecoratorHeight
					).intersection(adjustedCellBounds);
		case BOTTOM:
			return new Rectangle(
					adjustedCellBounds.x,
					adjustedCellBounds.y + adjustedCellBounds.height - preferredDecoratorHeight,
					adjustedCellBounds.width,
					preferredDecoratorHeight
					).intersection(adjustedCellBounds);
		case NONE:
		    break;
		}
		
		return null;
	}
	
	public ICellPainter getCellPainterAt(int x, int y, LayerCell cell, GC gc, Rectangle adjustedCellBounds, IConfigRegistry configRegistry) {
		if (getDecoratorCellPainterBounds(cell, gc, adjustedCellBounds, configRegistry).contains(x, y)) {
			return decoratorCellPainter.getCellPainterAt(x, y, cell, gc, adjustedCellBounds, configRegistry);
		} else if (getBaseCellPainterBounds(cell, gc, adjustedCellBounds, configRegistry).contains(x, y)) {
			return baseCellPainter.getCellPainterAt(x, y, cell, gc, adjustedCellBounds, configRegistry);
		} else {
			return this;
		}
	}
}
